/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import java.text.MessageFormat;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;

import de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlProgAntwort;
import de.bsvrz.iav.gllib.gllib.domain.comm.PrognoseAnfrage;
import de.bsvrz.iav.glprog.glprog.domain.comm.PrognoseAntwortSender;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Ein Worker, der Ganglinienprognoseanfragen bearbeitet.
 * <p>
 * Für zyklische Prognossn, muss die Prognose geprüft werden, wenn sich
 * Ereignisse ändern. Da Ereignisse nicht pro Messquerschnitt überwacht werden
 * können, wird die Menge aller Ereignisse auf Änderung überwacht und ggf. alle
 * zyklischen Prognosen überprüft.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public class AnfrageInBearbeitung implements Runnable {

    /** Der Logger der Klasse. */
    private final Debug log = Debug.getLogger();

    /** Die zu bearbeitende Anfrage. */
    private final GlProgAnfrage anfrage;

    /** Die zu erstellenden Prognoseganglinien. */
    private final List<GanglinieInBearbeitung> zyklischeGanglinien;

    /** Flag, ob sich Ereignisse geändert haben. */
    private boolean neueEreignisse;

    private GlProg applikation;

    private Consumer<Boolean> ereignisMengenHandler = (v) -> setNeueEreignisse(true);

    private PrognoseAntwortSender antwortSender;

    /**
     * Versorgt den Worker mit allen notwendigen Informationen, damit er loslegen
     * kann.
     *
     * @param applikation   die Ganglinienprognose-Applikation
     * @param anfrage       die Anfrage, die der Worker bearbeiten soll.
     * @param antwortSender der Sender zum Übermitteln der ermittelten Ganglinie
     */
    public AnfrageInBearbeitung(GlProg applikation, final GlProgAnfrage anfrage, PrognoseAntwortSender antwortSender) {
        zyklischeGanglinien = Collections.synchronizedList(new ArrayList<GanglinieInBearbeitung>());

        this.applikation = applikation;
        this.anfrage = anfrage;
        this.antwortSender = antwortSender;
    }

    @Override
    public void run() {

        applikation.getEreignisKalender().ereignisMengeGeaendert().addHandler(ereignisMengenHandler);

        final long startZeitstempel = System.currentTimeMillis();
        log.info(() -> MessageFormat.format("Starte die Bearbeitung einer neuen Anfrage: {0}", anfrage));

        List<MqGanglinie> prognosen = new ArrayList<>();

        // Alle Anfragen abarbeiten und die zyklischen Anfragen
        // zwischenspeichern
        for (final PrognoseAnfrage a : anfrage.getAnfragen()) {
            if ((a.getPrognoseZeitraum() == null) || !a.getPrognoseZeitraum().isValid()) {
                String message = MessageFormat.format(
                        "Die Anfrage kann nicht bearbeitet werden, da das Prognoseintervall ungültig ist: {0}", a);
                log.warning(message);
                applikation.warnung(a.getMessQuerschnitt(), message);
                continue;
            }

            final GanglinieInBearbeitung job = new GanglinieInBearbeitung(applikation, a);
            MqGanglinie prognostizierteGanglinie = job.prognostiziere();

            if (prognostizierteGanglinie == null) {
            	job.close();
                throw new IllegalStateException("Prognose wird bereits ausgeführt");
            }
            prognosen.add(prognostizierteGanglinie);
            if (a.isZyklischePrognose()) {
                zyklischeGanglinien.add(job);
            } else {
            	job.close();
            }
        }

        GlProgAntwort antwort = new GlProgAntwort(anfrage.getAnfrager(), anfrage.getAbsenderZeichen(), prognosen);
        antwortSender.sendeAntwort(antwort);

        log.info(() -> MessageFormat.format("InfoAnfrageFertig = Anfrage fertig bearbeitet: {0}, {1} Ganglinien", anfrage,
                antwort.getGanglinien().size()));

        final long endeZeitstempel = System.currentTimeMillis();
        final long berechnungsdauer = endeZeitstempel - startZeitstempel;
        if (TimeUnit.MILLISECONDS.toSeconds(berechnungsdauer) > 0) {
            log.info(() -> MessageFormat.format("Anfrage beantwortet in {0} Sekunden für {1}",
                    berechnungsdauer / TimeUnit.SECONDS.toMillis(1), anfrage));
        } else {
            log.info(() -> MessageFormat.format("Anfrage beantwortet in {0} Millisekunden für {1}", berechnungsdauer,
                    anfrage));
        }

        if (zyklischeGanglinien.isEmpty()) {
            beendeAnfrage();
        } else {
            applikation.getZyklischeAnfragenBearbeiter().addAnfrage(this);
        }

    }

    private void beendeAnfrage() {
        applikation.getEreignisKalender().ereignisMengeGeaendert().removeHandler(ereignisMengenHandler);
        antwortSender.abmelden();
    }

    /**
     * Gibt den Wert des Flags für Ereignisänderung zurück.
     *
     * @return {@code true}, wenn sich Ereignisse geändert haben.
     */
    public synchronized boolean isNeueEreignisse() {
        return neueEreignisse;
    }

    /**
     * Setzt das Flag für Ereignisänderung.
     *
     * @param neueEreignisse {@code true}, wenn sich Ereignisse geändert haben.
     */
    public synchronized void setNeueEreignisse(final boolean neueEreignisse) {
        this.neueEreignisse = neueEreignisse;
    }

    public boolean zyklischePrognoseAusfuehren() {

        if (!antwortSender.isSendeBereit()) {
            String message = MessageFormat.format(
                    "Der Sender der Anfrage ist nicht mehr vorhanden, die Antwort zur Prognoseanfrage wird verworfen: {0}",
                    anfrage);
            log.warning(message);
            applikation.warnung(null, message);

            // Aufräumen
            for (final GanglinieInBearbeitung g : zyklischeGanglinien) {
                g.close();
            }

            beendeAnfrage();
            return false;
        }

        List<MqGanglinie> zyklischePrognose = new ArrayList<>();
        Iterator<GanglinieInBearbeitung> iterator = zyklischeGanglinien.iterator();
        while (iterator.hasNext()) {
            final GanglinieInBearbeitung g = iterator.next();
            if (!g.isZyklischAktiv()) {
                g.close();
                iterator.remove();
                continue;
            }

            final LocalDateTime zeit = applikation.getTime();
            final boolean aktualisieren = zeit.isAfter(g.getNaechsteAktualisierung());
            if (zeit.isAfter(g.getNaechstePruefung()) || isNeueEreignisse() || g.isGanglinienModellPrognoseChanged()
                    || aktualisieren) {
                // Prognoseganglinie muss neu bestimmt werden

                log.fine(() -> MessageFormat.format("Zyklische Prognose muss geprüft werden für {0}", g.getAnfrage()));

                final MqGanglinie letztePrognose = g.getLetzteGueltigePrognoseganglinie();
                MqGanglinie prognostizierteGanglinie = g.prognostiziere();
                if (prognostizierteGanglinie != null) {
                    setNeueEreignisse(false);
                    final double abstand = GanglinienMQOperationen.basisabstand(g.getLetzteGueltigePrognoseganglinie(),
                            letztePrognose);
                    log.finer(() -> MessageFormat.format("Abstand zwischen letzter und aktueller Prognose: {1} - {0}",
                            g.getAnfrage(), abstand));

                    if (aktualisieren || (abstand > g.getAnfrage().getSchwelle())) {
                        // Prognoseganglinie muss gesendet werden
                        log.fine(() -> MessageFormat.format("Zyklische Prognose muss aktualisiert werden für {0}",
                                g.getAnfrage()));
                        zyklischePrognose.add(g.getLetzteGueltigePrognoseganglinie());
                    } else {
                        log.fine(() -> MessageFormat.format(
                                "Zyklische Prognose liegt unterhalb der Schwelle für die Aktualisierung für {0}",
                                g.getAnfrage()));
                    }
                }
            }
        }

        // Zyklische Prognose senden
        if (zyklischeGanglinien.isEmpty()) {
            beendeAnfrage();
            return false;
        }

        if (!zyklischePrognose.isEmpty()) {
            antwortSender.sendeAntwort(
                    new GlProgAntwort(anfrage.getAnfrager(), anfrage.getAbsenderZeichen(), zyklischePrognose));
        }
        
        return true;
    }
}
