/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import de.bsvrz.iav.gllib.gllib.domain.EreignisTyp;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;

/**
 * Fasst Ganglinien zusammen, die zu einem gemeinsamen Ereignistyp gehören,
 * welches entweder ein Tagesereignis ist oder nicht. Handelt es sich um ein
 * Tagesereignis, dann sind alle Ganglinien absolut, sonst sind sie alle relativ
 * (additiv oder multiplikativ).
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public class EreignisTypGanglinien {

	/** Der Ereignistyp, auf das sich alle enthaltenen Ganglinien beziehen. */
	private final EreignisTyp ereignisTyp;

	/** Flag ob es sich hier um Tagesganglinien handelt. */
	private boolean tagesganglinie;

	private final List<MqGanglinie> ganglinien = new ArrayList<>();

	/**
	 * Konstruiert die Ganglinienliste, in dem das Bezugsereignis festgelegt wird.
	 *
	 * @param ereignisTyp der Ereignistyp, worauf sich alle Ganglinien der Liste
	 *                    beziehen.
	 */
	public EreignisTypGanglinien(final EreignisTyp ereignisTyp) {
		assert ereignisTyp != null : "Ereignistyp darf nicht null sein.";
		this.ereignisTyp = ereignisTyp;
	}

	public EreignisTypGanglinien(EreignisTypGanglinien src) {
		ereignisTyp = src.getEreignisTyp();
		tagesganglinie = src.isTagesganglinie();
		ganglinien.addAll(src.ganglinien);
	}

	/**
	 * Gibt das Bezugsereignis der Ganglinienliste zurück.
	 *
	 * @return das Bezugsereignis.
	 */
	public EreignisTyp getEreignisTyp() {
		return ereignisTyp;
	}

	/**
	 * Gibt das Flag zurück, ob die Liste Ganglinien für ein Tagesereignis enthält.
	 *
	 * @return {@code true}, wenn die Ganglinien zu einem Tagesereignis gehören.
	 */
	public boolean isTagesganglinie() {
		return tagesganglinie;
	}

	/**
	 * Legt fest, ob es die Liste Ganglinien zu einem Tagesereignis sichert.
	 *
	 * @param tagesganglinie {@code true}, wenn die Ganglinien zu einem
	 *                       Tagesereignis gehören.
	 */
	public void setTagesganglinie(final boolean tagesganglinie) {
		this.tagesganglinie = tagesganglinie;
	}

	@Override
	public String toString() {
		String s;

		s = getClass().getSimpleName() + "[";
		s += "ereignisTyp=" + ereignisTyp;
		s += ", tagesganglinie=" + tagesganglinie;
		s += ", Ganglinienanzahl=" + ganglinien.size();

		return s + "]";
	}

	public void keepOnlyReferenzGl() {
		Iterator<MqGanglinie> iterator = ganglinien.iterator();
		while (iterator.hasNext()) {
			if (!iterator.next().isReferenz()) {
				iterator.remove();
			}
		}
	}

	public boolean hasGanglinien() {
		return !ganglinien.isEmpty();
	}

	public int getGanglinienAnzahl() {
		return ganglinien.size();
	}

	public void keepOnlyLast() {
		while (ganglinien.size() > 1) {
			ganglinien.remove(0);
		}
	}

	public void keepOnlyMaxVerschmelzungen() {
		MqGanglinie ganglinie = null;
		for (MqGanglinie gl : ganglinien) {
			if (ganglinie == null || gl.getAnzahlVerschmelzungen() >= ganglinie.getAnzahlVerschmelzungen()) {
				ganglinie = gl;
			}
		}

		ganglinien.clear();
		if (ganglinie != null) {
			ganglinien.add(ganglinie);
		}
	}

	public List<MqGanglinie> getGanglinien() {
		// TODO Unmodifiable?
		return ganglinien;
	}

	public void addGanglinie(MqGanglinie ganglinie) {
		ganglinien.add(ganglinie);
		
	}

}
