/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog;

import static de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen.addiere;
import static de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen.auschneiden;
import static de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen.multipliziere;
import static de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen.verbinde;
import static de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen.verschiebe;

import java.text.MessageFormat;
import java.time.Duration;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Consumer;

import de.bsvrz.iav.gllib.gllib.domain.ApproximationTyp;
import de.bsvrz.iav.gllib.gllib.domain.Ereignis;
import de.bsvrz.iav.gllib.gllib.domain.EreignisKalenderAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.EreignisTyp;
import de.bsvrz.iav.gllib.gllib.domain.EreignisTypLernParameter;
import de.bsvrz.iav.gllib.gllib.domain.GanglinieTyp;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.domain.Intervall;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.Messwerte;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.PrognoseIntervall;
import de.bsvrz.iav.gllib.gllib.domain.ZustandsWechsel;
import de.bsvrz.iav.gllib.gllib.domain.comm.PrognoseAnfrage;
import de.bsvrz.iav.gllib.gllib.shared.LocalDateTimeKonverter;
import de.bsvrz.iav.glprog.glprog.domain.GanglinienDatenbank;
import de.bsvrz.iav.glprog.glprog.domain.GanglinienModellPrognose;
import de.bsvrz.iav.glprog.glprog.domain.GanglinienSpeicherException;
import de.bsvrz.iav.glprog.glprog.util.AnalyseganglinieHelper;
import de.bsvrz.iav.glprog.glprog.util.Shifter;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Kapselt alle Daten, die zur Bearbeitung einer Anfrage notwendig sind.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public class GanglinieInBearbeitung {

    /** Der Logger der Klasse. */
    private final Debug log = Debug.getLogger();

    /** Die zu bearbeitende Anfrage. */
    private final PrognoseAnfrage anfrage;

    /** Das Zeichen des Absenders der Prognoseanfrage. */
    private final String absenderZeichen;

    /** Die historischen Ganglinien am Messquerschnitt. */
    private List<MqGanglinie> historischeGanglinien;

    /** Die Antwort vom Kalender. */
    private List<ZustandsWechsel> kalenderantwort;

    /** Die Analyseganglinie, die aus dem Archiv gelesen wird. */
    private MqGanglinie analyseganglinie;

    /** Enthält die Teilintervalle des Prognosezeitraums und deren Ganglinien. */
    private List<PrognoseIntervallGanglinien> prognoseIntervallGanglinien;

    /** Die prognostizierte Ganglinie. */
    private MqGanglinie prognoseganglinie;

    /**
     * Flag, ob sich der ganglinienModellPrognose am Messquerschnitt geändert hat.
     */
    private boolean ganglinienModellPrognoseChanged;

    /** Zeitstempel, wann die nächste Prüfung der Prognose fällig ist. */
    private LocalDateTime naechstePruefung;

    /** Zeitstempel, wann die nächste Aktualisierung der Prognose fällig ist. */
    private LocalDateTime naechsteAktualisierung;

    private final Lock singleThreadLock = new ReentrantLock();

    private GanglinienModellPrognose ganglinienModellPrognose;

    private final GlProg applikation;

    private Consumer<MessQuerschnitt> parameterChangeHandler = (mq) -> updateGanglinienModellPrognoseChanged(mq);

    /**
     * Initialisierung die Bearbeitung der Prognoseanfrage.
     *
     * @param applikation die Ganglinienprognose-Applikation
     * @param anfrage     die zu bearbeitende Anfrage.
     */
    public GanglinieInBearbeitung(GlProg applikation, final PrognoseAnfrage anfrage) {
        this.applikation = applikation;
        this.anfrage = anfrage;
        absenderZeichen = anfrage.getMessQuerschnitt() + "/" + applikation.getTime();
        ganglinienModellPrognose = applikation.getOnlineDatenLeser()
                .getGanglinienModellPrognose(anfrage.getMessQuerschnitt());
        this.applikation.getOnlineDatenLeser().ganglinienModellPrognoseChanged().addHandler(parameterChangeHandler);
    }

    /**
     * Schließt alle offenen Ressourcen. Das Objekt darf danach nicht mehr verwendet
     * werden.
     */
    public void close() {
        this.applikation.getOnlineDatenLeser().ganglinienModellPrognoseChanged().removeHandler(parameterChangeHandler);
    }

    /**
     * Gibt die zu bearbeitende Anfrage zurück.
     *
     * @return die Anfrage.
     */
    public PrognoseAnfrage getAnfrage() {
        return anfrage;
    }

    /**
     * Gibt die Prognoseganglinie zurück.
     *
     * @return eine Ganglinie, niemals {@code null}.
     */
    public MqGanglinie getLetzteGueltigePrognoseganglinie() {
        if (prognoseganglinie != null) {
            return prognoseganglinie;
        }

        // Leere Ganglinie erstellen, wenn keine Prognose möglich war
        final MqGanglinie ganglinie = new MqGanglinie(getAnfrage().getMessQuerschnitt());
        if (getAnfrage().getPrognoseZeitraum() != null) {
            ganglinie.setPrognoseZeitraum(anfrage.getPrognoseZeitraum());
        } else {
            ganglinie.setPrognoseZeitraum(Intervall.of(System.currentTimeMillis(), System.currentTimeMillis()));
        }
        return ganglinie;
    }

    public MqGanglinie prognostiziere() {

        if (!singleThreadLock.tryLock()) {
            return null;
        }

        try {

            final long startZeitstempel = System.currentTimeMillis();
            LocalDateTime prognoseStart = applikation.getTime();
            boolean ok = true;

            log.fine(() -> MessageFormat.format("Starte die Anfragebearbeitung für eine Ganglinie: {0}", anfrage));

            // Eingangsdaten ermitteln
            bestimmeEingangsdaten();
            ganglinienModellPrognoseChanged = false;
            log.fine(() -> MessageFormat.format("Alle notwendigen Eingangsdaten für Prognose zusammengetragen für {0}",
                    anfrage));

            if (historischeGanglinien.isEmpty()) {
                ok = false;
                String message = MessageFormat.format(
                        "Die Ganglinienprognose ist nicht durchführbar am Messquerschnitt {0}. Es existieren keine historischen Ganglinien.",
                        anfrage.getMessQuerschnitt());
                log.warning(() -> message);
                applikation.warnung(anfrage.getMessQuerschnitt(), message);
            }

            if ((kalenderantwort == null) || kalenderantwort.isEmpty()) {
                ok = false;
                String message = MessageFormat.format(
                        "Die Ganglinienprognose ist nicht durchführbar am Messquerschnitt {0}. Der Kalender hat keine Ereignisse im Prognosezeitraum {1} geliefert.",
                        anfrage.getMessQuerschnitt(), anfrage.getPrognoseZeitraum());
                log.warning(() -> message);
                applikation.warnung(anfrage.getMessQuerschnitt(), message);
            }

            if (ganglinienModellPrognose.isNull()) {
                ok = false;
                String message = MessageFormat.format(
                        "Die Ganglinienprognose ist nicht durchführbar am Messquerschnitt {0}. Der Parameter der Ganglinienprognose ({1}) wurde am Messquerschnitt nicht gesetzt.",
                        anfrage.getMessQuerschnitt(), "atg.ganglinienModellPrognose");
                log.warning(() -> message);
                applikation.warnung(anfrage.getMessQuerschnitt(), message);

            }

            if (!ok) {
                return getLetzteGueltigePrognoseganglinie();
            }

            // Prognose vorbereiten
            aufbereitenZustandswechsel();
            bestimmeTeilintervalle();

            if ((prognoseIntervallGanglinien == null) || prognoseIntervallGanglinien.isEmpty()) {
                prognoseganglinie = null;
                String message = MessageFormat.format("Keine Ereignisse im Prognosezeitraum für {0}", anfrage);
                log.warning(() -> message);
                applikation.warnung(anfrage.getMessQuerschnitt(), message);

                return getLetzteGueltigePrognoseganglinie();
            }

            // Prognose durchführen
            bestimmeTagesganglinien();
            bestimmeErlaubteRelativeEreignisTypen();
            bestimmePrognoseGanglinien();
            prognoseganglinie = erzeugePrognoseganglinie();
            if (prognoseganglinie != null) {
                prognoseganglinie.setPrognoseZeitraum(anfrage.getPrognoseZeitraum());
            }
            log.finer(() -> MessageFormat.format("Prognoseganglinie {1} für {0}", anfrage, prognoseganglinie));

            historischeGanglinien = null;
            kalenderantwort = null;
            prognoseIntervallGanglinien = null;
            ganglinienModellPrognose = null;
            naechstePruefung = LocalDateTime.from(prognoseStart).plus(anfrage.getPruefIntervall());
            updateNaechsteAktualisierung(prognoseStart);
            log.fine(() -> MessageFormat.format("Ganglinie fertig prognostiziert für {0}", anfrage));

            final long endeZeitstempel = System.currentTimeMillis();
            final long berechnungsdauer = endeZeitstempel - startZeitstempel;
            if ((berechnungsdauer / TimeUnit.SECONDS.toMillis(1)) > 0) {
                log.finer(() -> MessageFormat.format("Ganglinie für Anfrage progonstiziert in {0} Sekunden für {1}",
                        berechnungsdauer / TimeUnit.SECONDS.toMillis(1), anfrage));
            } else {
                log.finer(
                        () -> MessageFormat.format("Ganglinie für Anfrage progonstiziert in {0} Millisekunden für {1}",
                                berechnungsdauer, anfrage));
            }

            return getLetzteGueltigePrognoseganglinie();
        } finally {
            singleThreadLock.unlock();
        }
    }

    private void updateNaechsteAktualisierung(LocalDateTime prognoseStart) {
        if (naechsteAktualisierung == null || !naechsteAktualisierung.isAfter(prognoseStart)) {
            naechsteAktualisierung = prognoseStart.plus(anfrage.getSendeIntervall());
        }
    }

    /**
     * Gibt den nächsten Zeitpunkt zurück, an dem die Prognose überprüft werden
     * muss.
     *
     * @return der Zeitstempel mit der Fälligkeit der Prognoseprüfung.
     */
    public LocalDateTime getNaechstePruefung() {
        return naechstePruefung;
    }

    /**
     * Gibt den nächsten Zeitpunkt zurück, an dem die Prognose aktualisiert werden
     * muss.
     *
     * @return der Zeitstempel mit der Fälligkeit der Prognoseaktualisierung.
     */
    public LocalDateTime getNaechsteAktualisierung() {
        return naechsteAktualisierung;
    }

    /**
     * Bereitet die Liste der Zustandswechsel auf, damit der Scanline-Algorithmus
     * aus {@link #bestimmeTeilintervalle()} funktioniert.
     * <p>
     * Am Anfang der Liste können Zustandswechsel fehlen, die den Beginn des
     * Ereignisses signalisieren, wenn das Ereignis vor Beginn des Anfragezeitraumes
     * begonnen hat.
     * <p>
     * Ereignisse die vor dem Anfragezeitraum enden, werden aus der Liste entfernt.
     * <p>
     * Da die Gültigkeit für Ereignisse zeitlich und verkehrlich unterschiedlich
     * ist, hier aber nur die verkehrliche Gültigkeit interessiert, werden unnötige
     * Zustandswechsel entfernt. So ein unnötiger Zustandswechsel ist z.&nbsp;B.
     * wenn zweimal auf einander der selbe Zustand für verkehrliche Gültigkeit
     * kommt, weil sich die zeitliche Gültigkeit geändert hat.
     */
    private void aufbereitenZustandswechsel() {
        Iterator<ZustandsWechsel> iterator;
        final Set<Ereignis> scanline;
        Set<ZustandsWechsel> beginnen;

        /*
         * Für endende, aber nicht beginnende Ereignisse, den Anfangszeitpunkt auf
         * Prognosebeginn setzen.
         */
        scanline = new HashSet<>();
        beginnen = new HashSet<>();
        iterator = kalenderantwort.iterator();
        while (iterator.hasNext()) {

            ZustandsWechsel z = iterator.next();
            if (z.isVerkehrlichGueltig()) {
                // Ereignis beginnt
                if (!z.getZeitpunkt().isBefore(anfrage.getPrognoseZeitraum().getBis())) {
                    /*
                     * Ereignis beginnt am Ende oder nach dem Prognoseintervall und kann deswegen
                     * ignoriert werden.
                     */
                    iterator.remove();
                } else {
                    scanline.add(z.getEreignis());
                }
            } else {
                if (scanline.contains(z.getEreignis())) {
                    // begonnenes Ereignis endet
                    scanline.remove(z.getEreignis());
                } else {
                    // Ereignis endet, ohne begonnen zu haben
                    if (!z.getZeitpunkt().isAfter(anfrage.getPrognoseZeitraum().getVon())) {
                        /*
                         * Ereignis endet vor dem Prognoseintervall und kann komplett ignoriert werden.
                         */
                        iterator.remove();
                    } else {
                        // Den Beginn des Ereignisses ergänzen
                        beginnen.add(ZustandsWechsel.of(anfrage.getPrognoseZeitraum().getVon(), z.getEreignis(), true,
                                true));
                    }
                }
                if (!z.getZeitpunkt().isBefore(anfrage.getPrognoseZeitraum().getBis())) {
                    /*
                     * Ereignis endet am Ende oder nach dem Prognoseintervall, der Zustandswechsel
                     * ist nicht relevant
                     */
                    iterator.remove();
                }
            }
        }
        for (final ZustandsWechsel z : beginnen) {
            kalenderantwort.add(0, z);
        }

        // Doppelte Zustandswechsel aus der Liste entfernen.
        scanline.clear();
        iterator = kalenderantwort.iterator();
        while (iterator.hasNext()) {

            ZustandsWechsel z = iterator.next();
            if (z.isVerkehrlichGueltig()) {
                if (scanline.contains(z.getEreignis())) {
                    // Doppelter Beginn
                    iterator.remove();
                } else {
                    scanline.add(z.getEreignis());
                }
            } else {
                if (scanline.contains(z.getEreignis())) {
                    scanline.remove(z.getEreignis());
                } else {
                    // Doppeltes Ende
                    iterator.remove();
                }
            }
        }
    }

    /**
     * Erstellt aus der Liste der Zustandswechsels aus der Kalenderantwort die
     * enhtsprechenden Ereignisintervalle. Es wird ein Scanline-Algorithmus
     * verwendet, der die Liste der Zustandswechsel "scannt".
     * <p>
     * Die Methode entspricht zum Teil der ereignistypabhängigen Auswahl von
     * Ganglinien laut Feinspezifikation. Da der Ereigniskalender in seiner Antwort
     * nicht zwischen Tagesereignissen und Nichttagesereignissen unterscheidet, wird
     * diese Auswahl sowohl für relative als auch für absolute Ganglinien
     * durchgeführt.
     * <p>
     * Da die priorisierte Auswahl von Tagesganglinien und die ereignistypabhängige
     * Auswahl von (relativen) Ganglinien voneinander unabhängig sind, stellt die
     * gleichzeitige Ausführung keine Verletzung des Algorithmus dar.
     * <p>
     * Am Ende der Methode ist ist die Liste der Teilintervalle gefüllt (falls
     * möglich).
     *
     * @see #prognoseintervalle
     */
    private void bestimmeTeilintervalle() {
        Map<EreignisTyp, EreignisTypGanglinien> ganglinien;
        Set<Ereignis> scanline;
        LocalDateTime start = null;

        ganglinien = pruefeGanglinien(historischeGanglinien);
        prognoseIntervallGanglinien = new ArrayList<>();

        // Die Liste der Ereignisytpen im aktuellen Abschnitt
        scanline = new HashSet<>();

        /*
         * Scanline auf der Zeitachse entlangschieben. Die Scanline wird immer beim
         * Tageswechsel anhalten, da angenommen wird, dass die Ereignisse Montag,
         * Dienstag usw. immer existieren.
         */
        for (final ZustandsWechsel z : kalenderantwort) {
            Ereignis ereignis;

            ereignis = z.getEreignis();

            if (scanline.isEmpty()) {
                // Dies ist der Beginn eines neuen Intervalls
                start = z.getZeitpunkt();
                scanline.add(ereignis);
                continue;
            }

            if (start == null) {
                continue;
            }

            if (start.isEqual(z.getZeitpunkt()) && z.isVerkehrlichGueltig()) {
                /*
                 * Das Ereignis beginnt zusammen mit denen in der Scanline, ergo kein neues
                 * Intervall notwendig.
                 */
                scanline.add(ereignis);
            } else if (start.isEqual(z.getZeitpunkt()) && !z.isVerkehrlichGueltig()) {
                /*
                 * Das Ereignis endet zu Beginn des Scanline-Abschnitts. Der Ereignistyp wird
                 * nur aus der Scanline entfernt, aber kein neues Intervall begonnnen, da es die
                 * Breite 0 hätte.
                 */
                scanline.remove(ereignis);
            } else if (!start.isEqual(z.getZeitpunkt())) {
                assert !start
                        .isAfter(z.getZeitpunkt()) : "Die zeitliche Reihenfolge der Zustände ist nicht fortlaufend.";

                /*
                 * Ein Ereignis beginnt oder endet innerhalb eines anderen. Es wird die Scanline
                 * als Teilintervall abgelegt.
                 */
                if (z.getZeitpunkt().isAfter(anfrage.getPrognoseZeitraum().getVon())) {
                    final PrognoseIntervallGanglinien progIntervall;
                    final LocalDateTime start1;

                    if (start.isBefore(anfrage.getPrognoseZeitraum().getVon())) {
                        /*
                         * start liegt vor dem Prognosezeitraum und muss korrigiert werden
                         */
                        start1 = anfrage.getPrognoseZeitraum().getVon();
                    } else {
                        start1 = start;
                    }
                    progIntervall = new PrognoseIntervallGanglinien(PrognoseIntervall.of(start1, z.getZeitpunkt()));
                    for (final Ereignis e : scanline) {
                        if (ganglinien.containsKey(e.getEreignisTyp())) {
                            progIntervall.addEreignistypGanglinie(new EreignisTypGanglinien(ganglinien.get(e.getEreignisTyp())));
                        } else {
                            String message = MessageFormat.format(
                                    "Am Messquerschnitt {0} existiert für den Ereignistyp {1} keine Ganglinie.",
                                    anfrage.getMessQuerschnitt(), e.getEreignisTyp());
                            log.warning(() -> message);
                            applikation.warnung(anfrage.getMessQuerschnitt(), message);
                        }
                    }
                    prognoseIntervallGanglinien.add(progIntervall);
                }
                if (z.isVerkehrlichGueltig()) {
                    // Ein neues Ereignis beginnt
                    scanline.add(ereignis);
                } else {
                    // Ein Ereignis der Scanline endet
                    scanline.remove(ereignis);
                }
                start = z.getZeitpunkt();
            } else {
                assert false : "Unbehandelter Zustandswechsel";
            }
        }

        /*
         * Falls die Scanline noch Ereignisse enthält, wird ein Intervall bis zum Ende
         * des Prognosezeitraums ergänzt.
         */
        if (!scanline.isEmpty() && start != null) {
            PrognoseIntervallGanglinien progIntervall;
            final LocalDateTime start1;

            if (start.isBefore(anfrage.getPrognoseZeitraum().getVon())) {
                /*
                 * start liegt vor dem Prognosezeitraum und muss korrigiert werden
                 */
                start1 = anfrage.getPrognoseZeitraum().getVon();
            } else {
                start1 = start;
            }
            progIntervall = new PrognoseIntervallGanglinien(
                    PrognoseIntervall.of(start1, anfrage.getPrognoseZeitraum().getBis()));
            for (final Ereignis e : scanline) {
                if (ganglinien.containsKey(e.getEreignisTyp())) {
                    progIntervall.addEreignistypGanglinie(new EreignisTypGanglinien(ganglinien.get(e.getEreignisTyp())));
                } else {
                    String message = MessageFormat.format(
                            "Am Messquerschnitt {0} existiert für den Ereignistyp {1} keine Ganglinie.",
                            anfrage.getMessQuerschnitt(), e.getEreignisTyp());
                    log.warning(() -> message);
                    applikation.warnung(anfrage.getMessQuerschnitt(), message);
                }
            }
            prognoseIntervallGanglinien.add(progIntervall);
        }

        log.finer(() -> MessageFormat.format("Prognoseintervalle {1} für {0}", anfrage, prognoseIntervallGanglinien));
    }

    /**
     * Führt die priorisierte Auswahl der Tagesganglinie pro Tag durch. Als Ergebnis
     * enthält die Liste der Teilintervalle pro Teilintervall nur noch ein
     * Tagesereignis. Die Teilintervalle eines Tages besitzen alle das selbe
     * Tagesereignis.
     * <p>
     * Die Methode setzt voraus, dass die Teilintervalle des Prognosezeitraums
     * bestimmt wurde.
     *
     * @see #bestimmeTeilintervalle()
     */
    private void bestimmeTagesganglinien() {
        EreignisTyp tagestyp = null;

        for (final PrognoseIntervallGanglinien intervall : prognoseIntervallGanglinien) {
            List<EreignisTyp> ereignisTypen;
 
            ereignisTypen = intervall.getEreignisTypenAbsolut();
            if (ereignisTypen.isEmpty()) {
                /*
                 * Es gibt keine absoluten Ganglinien im Teilintervall und kann deswegen nicht
                 * weiter betrachtet werden.
                 */
                intervall.clear();
                continue;
            }

            LocalDateTime kalender = intervall.getIntervall().getVon();
            if ((kalender.getHour() == 0) && (kalender.getMinute() == 0) && (kalender.getSecond() == 0)
                    && (kalender.getNano() == 0)) {
                // Intervall ist der Beginn eines neuen Tages
                tagestyp = null;
            }

            if (tagestyp == null) {
                /*
                 * Neuer aktueller Tagesereignistyp wird festgelegt und notwendige Parameter
                 * abgerufen.
                 */
                tagestyp = ereignisTypen.get(0);
            }
            assert ereignisTypen.contains(tagestyp);
            while (ereignisTypen.contains(tagestyp)) {
                ereignisTypen.remove(tagestyp);
            }

            intervall.removeEreignisTypen(ereignisTypen);
            
        
        }

        log.finer(() -> MessageFormat.format(
                "Prognoseintervalle nach priorisierter Auswahl der Tagesganglinien {1} für {0}", anfrage,
                prognoseIntervallGanglinien));
    }

    /**
     * Entfernt pro Teilintervall der Prognose alle Ereignistypen, die sich nicht
     * mit dem gewählten Tagesereignis verknüpfen lassen.
     * <p>
     * Die Methode setzt voraus, dass es nur noch ein Tagesereignis pro
     * Teilintervall der Prognose gibt.
     *
     * @see #bestimmeTagesganglinien()
     */
    private void bestimmeErlaubteRelativeEreignisTypen() {
        for (final PrognoseIntervallGanglinien intervall : prognoseIntervallGanglinien) {
            EreignisTypGanglinien tagesGl;
            Iterator<EreignisTypGanglinien> iterator;

            tagesGl = intervall.getTagesEreignisTypGanglinien();
            if (tagesGl == null) {
                // Im aktuellen Teilintervall gibt es keine Ganglinien.
                continue;
            }

            iterator = intervall.getEreignisTypGanglinienRelativ().iterator();
            while (iterator.hasNext()) {
                EreignisTypGanglinien relativeGl;
                EreignisTypLernParameter datum;

                /*
                 * Verwerfe alle relativen Ganglinien, die sich nicht auf das ausgewählte
                 * Tagesereignis beziehen können.
                 */
                relativeGl = iterator.next();
                datum = relativeGl.getEreignisTyp().getLernParameter();

                if (!datum.isNull()) {
                    Set<EreignisTyp> bezug = datum.getBezugsTypen();
                    if (!bezug.isEmpty() && !bezug.contains(tagesGl.getEreignisTyp())) {
                        iterator.remove();
                    }
                }
            }
        }

        log.finer(() -> MessageFormat.format(
                "Prognoseintervalle nach Entfernung nicht verwendbarer relativer Ganglinien {1} für {0}", anfrage,
                prognoseIntervallGanglinien));
    }

    /**
     * Bestimmt eine relative Ganglinie pro Ereignistyp, die zur Prognose
     * herangezogen wird. Im Zeitbereich vor dem Horizont der mittelfristigen
     * Prognose wird das Pattern-Matching zur Auswahl verwendet. Danach werden die
     * Auswahlverfahren der langfristigen Prognose (statistische Auswahl und
     * Referenzganglinie) verwendet. Die Methoden der langfristigen Prognose werden
     * auch angewandt, wenn das Pattern-Matching kein Ergebnis geliefert hat.
     * <p>
     * Ist der Horizont der mittelfristigen Prognose gleich 0, werden nur die
     * Auswahlmethode der langfristigen Prognose angewandt.
     */
    private void bestimmePrognoseGanglinien() {
        for (final PrognoseIntervallGanglinien intervall : prognoseIntervallGanglinien) {
            boolean patternMatching = false;

            EreignisTypGanglinien tagesGl = intervall.getTagesEreignisTypGanglinien();
            if (tagesGl == null) {
                // Im aktuellen Teilintervall gibt es keine Ganglinien.
                continue;
            }

            Duration mittelfristigePrognose;
            if (anfrage.isNurLangfristigeAuswahl()) {
                mittelfristigePrognose = Duration.ZERO;
            } else {
                mittelfristigePrognose = ganglinienModellPrognose.getPatternMatchingHorizont();
            }
            assert !mittelfristigePrognose.isNegative() && !mittelfristigePrognose.isZero();

            Duration horizont = Duration.between(intervall.getIntervall().getVon(),
                    anfrage.getPrognoseZeitraum().getVon());

            // Versuche mittelfristige Prognose
            if ((analyseganglinie != null) && analyseganglinie.hatMesswerte() && (!horizont.isNegative())
                    && (horizont.compareTo(mittelfristigePrognose) < 0)) {
                // Mittelfristige Prognose
                patternMatching = patternMatching(intervall);
            }

            if (horizont.compareTo(mittelfristigePrognose) >= 0 || !patternMatching) {

                for (final EreignisTypGanglinien ergGl : intervall.getEreignisTypGanglinien()) {
                    // Mache langfristige Prognose Langfristige Prognose
                    if (ganglinienModellPrognose.getAuswahlMethode() == GanglinienAuswahlmethode.REFERENZ_GANGLINIE) {
                        /*
                         * Auswahl nach Referenzganglinie. Es werden alle Ganglinien entfernt, die keine
                         * Referenzganglinie sind.
                         */
                        ergGl.keepOnlyReferenzGl();
            
                        if (!ergGl.hasGanglinien()) {
                            String message = MessageFormat.format(
                                    "Am Messquerschnitt {0} gibt es für den Ereignistyp {1} keine Referenzganglinie.",
                                    anfrage.getMessQuerschnitt(), ergGl.getEreignisTyp());
                            log.warning(() -> message);
                            applikation.warnung(anfrage.getMessQuerschnitt(), message);
                        }

                        /*
                         * Falls mehr als eine Referenzganglinie vorhanden ist, eine auswählen und
                         * Warnung ausgeben.
                         */
                        if (ergGl.getGanglinienAnzahl() > 1) {
                            String message = MessageFormat.format(
                                    "Am Messquerschnitt {0} gibt es für den Ereignistyp {1} mehr als eine Referenzganglinie.",
                                    anfrage.getMessQuerschnitt(), ergGl.getEreignisTyp());
                            log.warning(() -> message);
                            applikation.warnung(anfrage.getMessQuerschnitt(), message);
                        }
                        ergGl.keepOnlyLast();
                        
                    } else {
                        /*
                         * Auswahl nach statistischer Methode, ergo Auswahl der Ganglinie mit der
                         * höchsten Anzahl der Verschmelzungen
                         */
                        ergGl.keepOnlyMaxVerschmelzungen();
                    }
                }
            }
        }
        log.finer(() -> MessageFormat.format(
                "Prognoseintervalle nach der mittelfristiger und langfristiger Prognose {1} für {0}", anfrage,
                prognoseIntervallGanglinien));
    }

    /**
     * Führt das Pattern-Matching für ein Teilintervall aus.
     *
     * @param intervall das Intrvall für das das Pattern-Matching ausgeführt werden
     *                  soll.
     * @return {@code true}, wenn das Pattern Matching erfolgreich war.
     */
    private boolean patternMatching(final PrognoseIntervallGanglinien intervall) {
        log.finer(() -> MessageFormat.format("Führe das Pattern Matching im Intervall {1} für die Anfrage {0} durch.",
                anfrage, intervall));

        final List<MqGanglinie> listeMatching;
        EreignisTypGanglinien listeAbsolut;
        final List<EreignisTypGanglinien> listeRelativ;

        listeAbsolut = intervall.getTagesEreignisTypGanglinien();
        listeRelativ = intervall.getEreignisTypGanglinienRelativ();

        if (!listeAbsolut.hasGanglinien()) {
            throw new IllegalStateException(MessageFormat.format(
                    "Das Prognoseintervall enthält keine Tagesganglinien, es wird aber mindestens eine Tagesganglinie erwartet: {0}",
                    intervall));
        }

        // Vergleichsliste zusammenstellen
        listeMatching = new ArrayList<>();
        for (final MqGanglinie glAbsolut : listeAbsolut.getGanglinien()) {
            MqGanglinie ganglinie;

            if (!listeRelativ.isEmpty()) {
                final Shifter shifter;
                final int[] maximas;

                /*
                 * Der Shifter hält pro relativen Ereignis den Index der aktuellen Ganglinie
                 */
                maximas = new int[listeRelativ.size()];
                for (int i = 0; i < maximas.length; ++i) {
                    maximas[i] = listeRelativ.get(i).getGanglinien().size() - 1;
                }
                shifter = new Shifter(maximas);

                do {
                    final List<MqGanglinie> listeAdd, listeMul;

                    /*
                     * Bestimme für jedes relative Ereignis die aktuelle Ganglinie und füge sie der
                     * Liste der additiven bzw. multiplikativen Ganglinien hinzu
                     */
                    listeAdd = new ArrayList<>();
                    listeMul = new ArrayList<>();
                    for (int i = 0; i < shifter.getLength(); ++i) {
                        MqGanglinie g;

                        g = listeRelativ.get(i).getGanglinien().get(shifter.getValues()[i]);
                        if (g.getTyp() == GanglinieTyp.ADDITIV) {
                            listeAdd.add(g);
                        } else if (g.getTyp() == GanglinieTyp.MULTIPLIKATIV) {
                            listeMul.add(g);
                        } else {
                            throw new IllegalStateException(MessageFormat.format(
                                    "Die relative Ganglinie muss additiv oder multiplikativ sein, der Typ war aber {0}",
                                    g.getTyp()));
                        }
                    }

                    // Kombiniere die absolute mit den relativen Ganglinien
                    ganglinie = glAbsolut.erzeugeKopie();
                    for (final MqGanglinie g : listeMul) {
                        if (ganglinie.getApproximationDaK() == g.getApproximationDaK()) {
                            ganglinie = multipliziere(ganglinie, g);
                        }
                        // Bei unterschiedlicher Approximation, Ganglinie ignorieren
                    }
                    for (final MqGanglinie g : listeAdd) {
                        if (ganglinie.getApproximationDaK() == g.getApproximationDaK()) {
                            ganglinie = addiere(ganglinie, g);
                        }
                        // Bei unterschiedlicher Approximation, Ganglinie ignorieren
                    }

                    // Die bestimmte Ganglinien in Vergleichsliste aufnehmen
                    listeMatching.add(ganglinie);
                } while (shifter.shift());
            } else {
                listeMatching.add(glAbsolut.erzeugeKopie());
            }
        }

        /*
         * Archivganglinie auf 1.1.1970 verschieben, damit sie mit historischen
         * Ganglinien verglichen werden kann.
         */
        final MqGanglinie archivGl;
        final Calendar cal;
        final EreignisTypGanglinien result;
        final GanglinienMQOperationen.MqGangliniePatternMatchingErgebnis matchingErg;

        archivGl = analyseganglinie.erzeugeKopie();
        cal = Calendar.getInstance();
        cal.setTimeInMillis(archivGl.getStuetzstellen().get(0).getZeitstempel());
        cal.set(Calendar.HOUR_OF_DAY, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MILLISECOND, 0);
        GanglinienMQOperationen.verschiebe(archivGl, -cal.getTimeInMillis());

        // Eigentliches Pattern-Matching ausführen
        matchingErg = GanglinienMQOperationen.patternMatching(archivGl, listeMatching,
                ganglinienModellPrognose.getPatternMatchingOffset().toMillis(),
                ganglinienModellPrognose.getPatternMatchingOffset().toMillis(),
                ganglinienModellPrognose.getMatchingIntervall().toMillis());
        log.finer(() -> MessageFormat.format("Das Ergebnis des Pattern-Matching {1} für {0}", anfrage, matchingErg));

        if (matchingErg.getAbstand() > ganglinienModellPrognose.getMaxMatchingFehler()) {
            // Keine passende Ganglinie gefunden
            return false;
        }

        // Alle anderen Ganglinie im Intervall entfernen und Ergebnis sichern
        result = new EreignisTypGanglinien(listeAbsolut.getGanglinien().get(0).getEreignisTyp());
        result.setTagesganglinie(true);
        result.addGanglinie(matchingErg.getGanglinie());
        intervall.clear();
        intervall.addEreignistypGanglinie(result);
        return true;
    }

    /**
     * Prüft ob die Liste der Ganglinien pro Ereignistyp plausibel ist. Pro
     * Ereignistyp darf es entweder nur absolute oder nur relative Ganglinien geben.
     * Ist dies nicht der Fall, ist die Ganglinienliste für das betreffende Ereignis
     * ungültig und wird ignoriert. Für den betreffenden Ganglinientyp wird eine
     * Warnung ausgegeben.
     *
     * @param ganglinien alle verf&uuml;gbaren Ganglinien im Prognosezeitraum.
     * @return eine {@code Map} mit g&uuml;ltigen Ganglinien pro Ereignistyp.
     */
    private Map<EreignisTyp, EreignisTypGanglinien> pruefeGanglinien(final List<MqGanglinie> ganglinien) {
        final Map<EreignisTyp, EreignisTypGanglinien> result;
        final Set<EreignisTyp> ausschuss;

        result = new HashMap<>();
        ausschuss = new HashSet<>();
        // Alle Ganglinien durchlaufen und nach Ereignistyp in Ergebis sortieren
        for (final MqGanglinie gl : ganglinien) {
            EreignisTyp glTyp;
            EreignisTypGanglinien glListe;

            glTyp = gl.getEreignisTyp();
            if (ausschuss.contains(glTyp)) {
                /*
                 * Der Ereignistyp dieser Ganglinie wurde bereits als ungültig erkannt, ergo
                 * muss die Ganglinie nicht weiter untersucht werden.
                 */
                continue;
            }

            if (!result.containsKey(glTyp)) {
                /*
                 * Der Ereignistyp wird zum ersten Mal untersucht, ergo erstmal in Ergebnis mit
                 * leerer Ganglinienliste aufnehmen.
                 */
                result.put(glTyp, new EreignisTypGanglinien(glTyp));
            }
            glListe = result.get(glTyp);

            if (glListe.getGanglinienAnzahl() == 0) {
                /*
                 * Dies ist die erste Ganglinie des Ereignistyps, weswegen er den bestimmt ob
                 * die Ganglinien zu einem Tagesereignis gehören oder nicht.
                 */
                glListe.addGanglinie(gl);
                glListe.setTagesganglinie(gl.isAbsolut());
                continue;
            }

            if ((glListe.isTagesganglinie() && gl.isAbsolut()) || (!glListe.isTagesganglinie() && !gl.isAbsolut())) {
                /*
                 * Nach hinzufügen der Ganglinie würde die Liste des Ereignistyps nur absolute
                 * oder nur relative Ganglinien / enthalten.
                 */
                glListe.addGanglinie(gl);
            } else {
                /*
                 * Es gibt für den Ereignistyp absolute und relative Ganglinien. Dies ist ein
                 * inkonsistenter Zustand.
                 */
                ausschuss.add(glTyp);
                result.remove(glTyp);
                String message = MessageFormat.format(
                        "Am Messquerschnitt {0} liegen für den Ereignistyp {1} sowohl absolute als auch relative Ganglinien vor. Der Ereignistyp wird wegen des inkonsistenten Zustands nicht beachtet.",
                        anfrage.getMessQuerschnitt(), glTyp);
                log.warning(() -> message);
                applikation.warnung(anfrage.getMessQuerschnitt(), message);
            }
        }

        return result;
    }

    /**
     * Verschmilzt pro Prognoseintervall die realtiven Ganglinien mit der absoluten
     * Ganglinie.
     *
     * @return die Prognoseganglinie, f&uuml;r das gesamte Prognoseintervall.
     */
    private MqGanglinie erzeugePrognoseganglinie() {
        List<MqGanglinie> teile;
        MqGanglinie ganglinie = null;

        // Ganglinie pro Teilintervall bestimmen
        LocalDateTime tag = LocalDateTime.of(anfrage.getPrognoseZeitraum().getVon().toLocalDate(), LocalTime.MIDNIGHT);
        teile = new ArrayList<>();
        for (final PrognoseIntervallGanglinien intervall : prognoseIntervallGanglinien) {
            MqGanglinie intervallGl;
            List<MqGanglinie> addGl;
            List<MqGanglinie> mulGl;
            LocalDateTime teilIntervall;

            teilIntervall = intervall.getIntervall().getVon();
            if ((teilIntervall.getHour() == 0) && (teilIntervall.getMinute() == 0) && (teilIntervall.getSecond() == 0)
                    && (teilIntervall.getNano() == 0)) {
                tag = teilIntervall;
            }

            if ((intervall.getTagesEreignisTypGanglinien() == null)
                    || !intervall.getTagesEreignisTypGanglinien().hasGanglinien()) {
                // Es gibt keine Ganglinien im Teilintervall
                continue;
            }

            // Tagesganglinie im Teilintervall bestimmen
            assert intervall.getTagesEreignisTypGanglinien()
                    .getGanglinienAnzahl() == 1 : "Es darf wegen der vorangegangenen Auswahl nur noch eine Tagesganglinie geben.";
            intervallGl = intervall.getTagesEreignisTypGanglinien().getGanglinien().get(0).erzeugeKopie();

            // Additive relative Ganglinien bestimmen
            addGl = new ArrayList<>();
            for (final EreignisTypGanglinien typGl : intervall.getEreignisTypGanglinienRelativ()) {
                if (!typGl.hasGanglinien()) {
                    continue;
                }
                MqGanglinie gl = typGl.getGanglinien().get(0);
                if (gl.getTyp() == GanglinieTyp.ADDITIV) {
                    addGl.add(gl);
                }
            }

            // Multiplikative relative Ganglinien bestimmen
            mulGl = new ArrayList<>();
            for (final EreignisTypGanglinien typGl : intervall.getEreignisTypGanglinienRelativ()) {
                if (!typGl.hasGanglinien()) {
                    continue;
                }
                MqGanglinie gl = typGl.getGanglinien().get(0);
                if (gl.getTyp() == GanglinieTyp.MULTIPLIKATIV) {
                    mulGl.add(gl);
                }
            }

            // Tagesganglinie und relative Ganglinien verknüpfen
            for (final MqGanglinie g : mulGl) {
                if (intervallGl.getApproximationDaK() == g.getApproximationDaK()) {
                    intervallGl = multipliziere(intervallGl, g);
                } // Bei unterschiedlicher Approximation, Ganglinie ignorieren
            }
            for (final MqGanglinie g : addGl) {
                if (intervallGl.getApproximationDaK() == g.getApproximationDaK()) {
                    intervallGl = addiere(intervallGl, g);
                } // Bei unterschiedlicher Approximation, Ganglinie ignorieren
            }

            // Spezeilbehandlung der Stützstellen für die Tage an denen die
            // Sommerzeitunschaltung erfolgt. */
            adaptiereGanglinieFuerSommerZeitWechsel(intervallGl, tag);

            verschiebe(intervallGl, LocalDateTimeKonverter.toMillis(tag));
            auschneiden(intervallGl, Intervall.of(intervall.getIntervall()));
            teile.add(intervallGl);
        }

        // Ganglinien der Teilintervalle verketten
        for (final MqGanglinie g : teile) {
            if (ganglinie == null) {
                ganglinie = g;
                ganglinie.setApproximationDaK(g.getApproximationDaK());
            } else {
                ganglinie = verbinde(ganglinie, g, Long.MAX_VALUE);
                if (ganglinie.getApproximationDaK() != g.getApproximationDaK()) {
                    ganglinie.setApproximationDaK(ApproximationTyp.UNBESTIMMT);
                }
            }
        }

        return ganglinie;
    }

    private void adaptiereGanglinieFuerSommerZeitWechsel(MqGanglinie gl, LocalDateTime tag) {

        LocalDateTime tagEnde = LocalDateTime.of(tag.toLocalDate().plusDays(1), LocalTime.MIDNIGHT);

        long stunden = Duration.between(tag, tagEnde).toHours();
        if (stunden == 23) {

            Map<Long, Messwerte> adaptierteWerte = new TreeMap<>();
            Set<Long> removed = new LinkedHashSet<>();
            for (Entry<Long, Messwerte> value : gl.getMesswerte().entrySet()) {
                if (value.getKey() >= TimeUnit.HOURS.toMillis(2)) {
                    if (value.getKey() >= TimeUnit.HOURS.toMillis(3)) {
                        adaptierteWerte.put(value.getKey() - TimeUnit.HOURS.toMillis(1), value.getValue());
                    }
                    removed.add(value.getKey());
                }
            }
            gl.loescheMesswerte(removed);
            gl.aktualisiereMesswerte(adaptierteWerte);

        } else if (stunden == 25) {

            Set<Long> removed = new LinkedHashSet<>();
            Map<Long, Messwerte> adaptierteWerte = new TreeMap<>();
            for (Entry<Long, Messwerte> value : gl.getMesswerte().entrySet()) {
                if (value.getKey() > TimeUnit.HOURS.toMillis(2)) {
                    adaptierteWerte.put(value.getKey() + TimeUnit.HOURS.toMillis(1), value.getValue());
                    removed.add(value.getKey());
                }
            }
            gl.loescheMesswerte(removed);
            gl.aktualisiereMesswerte(adaptierteWerte);

        } else if (stunden != 24) {
            throw new IllegalArgumentException("Tage mit " + stunden + " Stunden werden nicht erwartet und behandelt!");
        }
    }

    /**
     * Bestimmt alle notwendigen Eingangsdaten der Prognose.
     */
    private void bestimmeEingangsdaten() {
        ganglinienModellPrognose = applikation.getOnlineDatenLeser()
                .getGanglinienModellPrognose(anfrage.getMessQuerschnitt());
        log.finer(() -> MessageFormat.format("ganglinienModellPrognose für {0} ist {1}", anfrage.getMessQuerschnitt(),
                ganglinienModellPrognose));

        if (!getAnfrage().isNurLangfristigeAuswahl()) {
            final AnalyseganglinieHelper analyseganglinieHelper = new AnalyseganglinieHelper(applikation, getAnfrage());
            analyseganglinieHelper.setPrognoseParameter(ganglinienModellPrognose);
            analyseganglinie = analyseganglinieHelper.getAnalyseganglinie();
        }

        historischeGanglinien = getHistorischeGanglinien();
        log.finest(() -> MessageFormat.format("Historische Ganglinien am Messquerschnitt {0}: {1}",
                anfrage.getMessQuerschnitt(), historischeGanglinien));
        kalenderantwort = getEreignisse();
    }

    /**
     * Prüft, ob dies die Bearbeitung einer zyklischen Prognoseanfrage ist und ob
     * sie noch gültig ist. Eine zyklische Prognose ist bis zum Ende des angefragten
     * Prognosezeitraums gültig.
     *
     * @return {@code true}, wenn dies eine zyklische Prognose ist und der
     *         Prognosezeitraum noch nicht abgelaufen ist.
     */
    public boolean isZyklischAktiv() {
        return anfrage.isZyklischePrognose() && (anfrage.getPrognoseZeitraum().getBis().isAfter(applikation.getTime()));
    }

    @Override
    public String toString() {
        String s;

        s = getClass() + "[";
        s += "absenderZeichen=" + absenderZeichen;
        s += ", naechstePruefung=" + LocalDateTimeKonverter.formatierteZeit(naechstePruefung);
        s += ", naechsteAktualisierung=" + LocalDateTimeKonverter.formatierteZeit(naechsteAktualisierung);
        s += ", anfrage=" + anfrage;
        s += "]";

        return s;
    }

    /**
     * Gibt den Wert des Flags für die Änderung des ganglinienModellPrognoses am
     * Messquerschnitt zurück.
     *
     * @return {@code true}, wenn sich der Parameter geändert hat.
     */
    public boolean isGanglinienModellPrognoseChanged() {
        return ganglinienModellPrognoseChanged;
    }

    /**
     * Setzt das Flag für die Änderung des ganglinienModellPrognoses am
     * Messquerschnitt.
     *
     * @param mq {@code true}, wenn sich der Parameter geändert hat.
     */
    private void updateGanglinienModellPrognoseChanged(final MessQuerschnitt mq) {
        if (anfrage.getMessQuerschnitt().equals(mq)) {
            this.ganglinienModellPrognoseChanged = true;
        }
    }

    /**
     * Bezieht die historischen Ganglinien des Messquerschnitts.
     *
     * @return die historischen Ganglinien, aber niemals <code>null</code>.
     */
    private List<MqGanglinie> getHistorischeGanglinien() {
        GanglinienDatenbank db = applikation.getGanglinienDatenbank();
        try {
            return db.leseGanglinien(anfrage.getMessQuerschnitt(), anfrage.getEreignisTypen());
        } catch (GanglinienSpeicherException e) {
            log.warning(() -> "Ganglinien für MQ '" + anfrage.getMessQuerschnitt() + "' konnten nicht gelesen werden: "
                    + e.getLocalizedMessage());
        }

        return Collections.emptyList();
    }

    private List<ZustandsWechsel> getEreignisse() {

        EreignisKalenderAnfrage leseAnfrage = new EreignisKalenderAnfrage(absenderZeichen);
        leseAnfrage.setIntervall(Intervall.of(anfrage.getPrognoseZeitraum()));
        anfrage.getEreignisTypen().forEach(t -> leseAnfrage.addEreignisTyp(t));
        leseAnfrage.addRaeumlicheGueltigkeit(anfrage.getMessQuerschnitt().getStrassenSegmentPid());
        try {
            return new ArrayList<>(applikation.getEreignisKalender()
                    .kalenderAnfragen(leseAnfrage, Duration.ofMinutes(5)).getZustandsWechsel());
        } catch (TimeoutException e) {
            log.warning(() -> e.getLocalizedMessage());
            return Collections.emptyList();
        }
    }
}
