/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog.portals;

import java.util.ArrayList;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.DataNotSubscribedException;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.Messwerte;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.Stuetzstelle;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSpeicherAntwort;
import de.bsvrz.iav.gllib.gllib.shared.LocalDateTimeKonverter;
import de.bsvrz.iav.glprog.glprog.domain.comm.GanglinienAntwortSender;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Implementierung eines {@link GanglinienAntwortSender} zum Versenden von
 * Antworten auf Anfragen an den Speicher historischer Ganglinien.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class DavGanglinienAntwortSender implements GanglinienAntwortSender {

    private static final Debug LOGGER = Debug.getLogger();

    private static ClientSenderInterface sender = new ClientSenderInterface() {

        @Override
        public boolean isRequestSupported(SystemObject object, DataDescription dataDescription) {
            return false;
        }

        @Override
        public void dataRequest(SystemObject object, DataDescription dataDescription, byte state) {
            // Sendesteuerung wird nicht ausgewertet
        }
    };

    private final ClientDavInterface dav;
    private final AttributeGroup anfrageAtg;
    private final Aspect anfrageAsp;

    private SystemObject speicherObj;

    public DavGanglinienAntwortSender(ClientDavInterface dav, SystemObject speicherObj) {
        this.dav = dav;
        anfrageAtg = dav.getDataModel().getAttributeGroup("atg.ganglinienSpeicherAntwort");
        anfrageAsp = dav.getDataModel().getAspect("asp.antwort");
        this.speicherObj = speicherObj;
        try {
            dav.subscribeSender(sender, speicherObj, new DataDescription(anfrageAtg, anfrageAsp), SenderRole.source());
        } catch (OneSubscriptionPerSendData e) {
            LOGGER.warning(e.getLocalizedMessage());
        }
    }

    @Override
    public void anfrageBeantworten(GlSpeicherAntwort antwort) {
        try {
            dav.sendData(konvertiereAntwort(antwort));
        } catch (DataNotSubscribedException | SendSubscriptionNotConfirmed e) {
            LOGGER.warning(e.getLocalizedMessage());
        }

    }

    private ResultData konvertiereAntwort(GlSpeicherAntwort antwort) {

        Data daten = dav.createData(anfrageAtg);

        daten.getReferenceValue("absenderId")
                .setSystemObject(dav.getDataModel().getObject(antwort.getAnfrager().getId()));
        daten.getTextValue("AbsenderZeichen").setText(antwort.getAbsenderZeichen());

        Data antwortDaten = daten.getItem("Antwort");
        antwortDaten.getReferenceValue("Messquerschnitt").setSystemObjectPid(antwort.getMessQuerschnitt().getPid());

        if (antwort.getFehlerMeldung() != null) {
            antwortDaten.getArray("Fehlermeldung").setLength(1);
            antwortDaten.getArray("Fehlermeldung").getTextValue(0).setText(antwort.getFehlerMeldung());
        }

        List<MqGanglinie> versendbareGanglinien = new ArrayList<>();
        for (MqGanglinie ganglinie : antwort.getGanglinien()) {
            SystemObject ereignisTypObject = dav.getDataModel().getObject(ganglinie.getEreignisTyp().getPid());
            if (ereignisTypObject != null) {
                versendbareGanglinien.add(ganglinie);
            }
        }
        
        Array ganglinienArray = antwortDaten.getArray("GespeicherteGanglinien");
        ganglinienArray.setLength(versendbareGanglinien.size());

        int glIdx = 0;
        for (MqGanglinie ganglinie : versendbareGanglinien) {
            Data glData = ganglinienArray.getItem(glIdx++);
            glData.getReferenceValue("EreignisTyp").setSystemObjectPid(ganglinie.getEreignisTyp().getPid());
            glData.getUnscaledValue("GanglinieId").set(ganglinie.getGanglinieId());
            glData.getUnscaledValue("Referenzganglinie").setText(ganglinie.isReferenz() ? "Ja" : "Nein");
            glData.getUnscaledValue("AnzahlVerschmelzungen").set(ganglinie.getAnzahlVerschmelzungen());
            glData.getTimeValue("LetzteVerschmelzung")
                    .setMillis(LocalDateTimeKonverter.toMillis(ganglinie.getLetzteVerschmelzung()));
            glData.getUnscaledValue("GanglinienTyp").set(ganglinie.getTyp().getCode());
            glData.getUnscaledValue("GanglinienVerfahren").set(ganglinie.getApproximationDaK().getCode());
            glData.getUnscaledValue("Ordnung").set(ganglinie.getBSplineOrdnung());

            Array ssArray = glData.getArray("Stützstelle");
            ssArray.setLength(ganglinie.getStuetzStellenAnzahl());

            int ssIdx = 0;
            for (Stuetzstelle<Messwerte> stuetzstelle : ganglinie.getStuetzstellen()) {
                Data ssData = ssArray.getItem(ssIdx++);
                ssData.getTimeValue("Zeit").setMillis(stuetzstelle.getZeitstempel());
                ssData.getScaledValue("QLkw").set(stuetzstelle.getWert().getQLkwNichtNull());
                ssData.getScaledValue("QKfz").set(stuetzstelle.getWert().getQKfzNichtNull());
                ssData.getScaledValue("VPkw").set(stuetzstelle.getWert().getVPkwNichtNull());
                ssData.getScaledValue("VLkw").set(stuetzstelle.getWert().getVLkwNichtNull());
            }
        }

        return new ResultData(speicherObj, new DataDescription(anfrageAtg, anfrageAsp), dav.getTime(), daten);
    }
}
