/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog.portals;

import java.time.Duration;
import java.util.ArrayList;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.Array;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienAnfrager;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.PrognoseIntervall;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlProgAnfrage;
import de.bsvrz.iav.gllib.gllib.domain.comm.PrognoseAnfrage;
import de.bsvrz.iav.gllib.gllib.portals.DavObjektFactory;
import de.bsvrz.iav.gllib.gllib.shared.Event;
import de.bsvrz.iav.glprog.glprog.domain.comm.PrognoseAnfrageEmpfaenger;

/**
 * 
 * Implementierung des {@link PrognoseAnfrageEmpfaenger} zum Empfang von
 * Anfragen an die Ganglinienprognose über die Datenverteilerschnittstelle.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class DavPrognoseAnfrageEmpfaenger implements PrognoseAnfrageEmpfaenger {

    private final ClientReceiverInterface receiver = new ClientReceiverInterface() {

        @Override
        public void update(ResultData[] results) {
            for (ResultData result : results) {
                if (result.hasData()) {
                    empfangenEvent.send(konvertiere(result.getData()));
                }
            }
        }
    };

    private final Event<GlProgAnfrage> empfangenEvent = new Event<>();
    private final DavObjektFactory objektFactory;

    public DavPrognoseAnfrageEmpfaenger(DavObjektFactory objektFactory) {

        this.objektFactory = objektFactory;
        ClientDavInterface dav = objektFactory.getDav();

        AttributeGroup atg = dav.getDataModel().getAttributeGroup("atg.prognoseGanglinienAnfrage");
        Aspect asp = dav.getDataModel().getAspect("asp.anfrage");
        dav.subscribeReceiver(receiver, dav.getLocalConfigurationAuthority(), new DataDescription(atg, asp),
                ReceiveOptions.normal(), ReceiverRole.drain());
    }

    private GlProgAnfrage konvertiere(Data daten) {

        SystemObject anfragerObject = daten.getReferenceValue("absenderId").getSystemObject();
        GanglinienAnfrager anfrager = objektFactory.createAnfrager(anfragerObject.getId());

        String absenderzeichen = daten.getTextValue("AbsenderZeichen").getText();
        List<PrognoseAnfrage> anfragen = new ArrayList<>();

        Array anfragenArray = daten.getArray("PrognoseGanglinienAnfrage");
        for (int i = 0; i < anfragenArray.getLength(); ++i) {

            Data anfrageItem = anfragenArray.getItem(i);

            SystemObject mqObject = anfrageItem.getReferenceValue("Messquerschnitt").getSystemObject();
            MessQuerschnitt mq = objektFactory.createMessQuerschnitt(mqObject);

            PrognoseIntervall prognoseZeitraum = PrognoseIntervall.of(anfrageItem.getTimeValue("ZeitpunktPrognoseBeginn").getMillis(),
                    anfrageItem.getTimeValue("ZeitpunktPrognoseEnde").getMillis());
            boolean nurLangfristigeAuswahl = anfrageItem.getUnscaledValue("NurLangfristigeAuswahl").getState().getName()
                    .equalsIgnoreCase("ja");

            PrognoseAnfrage anfrage = new PrognoseAnfrage(mq, prognoseZeitraum, nurLangfristigeAuswahl);

            Array ereignisTypArray = anfrageItem.getArray("EreignisTyp");
            for (int ereignisTypIdx = 0; ereignisTypIdx < ereignisTypArray.getLength(); ++ereignisTypIdx) {
                SystemObject etObject = ereignisTypArray.getReferenceValue(ereignisTypIdx).getSystemObject();
                if (etObject != null) {
                    anfrage.addEreignisTyp(objektFactory.createEreignisTyp(etObject.getPid()));
                }
            }

            anfrage.setZyklischePrognose(
                    anfrageItem.getUnscaledValue("ZyklischePrognose").getState().getName().equalsIgnoreCase("ja"));
            anfrage.setPruefIntervall(Duration.ofSeconds(anfrageItem.getUnscaledValue("Überprüfungsintervall").longValue()));
            anfrage.setSchwelle(anfrageItem.getUnscaledValue("Aktualisierungsschwelle").intValue());
            anfrage.setSendeIntervall(Duration.ofMillis(anfrageItem.getUnscaledValue("Aktualisierungsintervall").longValue()));
            anfragen.add(anfrage);
        }

        return new GlProgAnfrage(anfrager, absenderzeichen, anfragen);
    }

    @Override
    public Event<GlProgAnfrage> anfrageEmpfangen() {
        return empfangenEvent;
    }

}
