/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.1 Ganglinienprognose
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.glprog.glprog.portals;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.time.Duration;
import java.time.LocalTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllib.gllib.domain.ApproximationTyp;
import de.bsvrz.iav.gllib.gllib.domain.EreignisTyp;
import de.bsvrz.iav.gllib.gllib.domain.GanglinieTyp;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.Messwerte;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.ObjektFactory;
import de.bsvrz.iav.gllib.gllib.domain.Stuetzstelle;
import de.bsvrz.iav.gllib.gllib.modell.db.DbGanglinie;
import de.bsvrz.iav.gllib.gllib.portals.DavObjektFactory;
import de.bsvrz.iav.gllib.gllib.shared.LocalDateTimeKonverter;
import de.bsvrz.iav.glprog.glprog.domain.GanglinienDatenbank;
import de.bsvrz.iav.glprog.glprog.domain.GanglinienSpeicherException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Gangliniendatenbank-Implementierung mit einer eingebetten Derby-Datenbank.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class DerbyGanglinienDatenbank implements GanglinienDatenbank {

    private static final Debug LOGGER = Debug.getLogger();
    private final String dbDir;
    private final ObjektFactory objektFactory;

    private Object dbLock = new Object();

    public DerbyGanglinienDatenbank(ObjektFactory objektFactory, String dbDir, LocalTime compressTime) {
        this.dbDir = dbDir;
        this.objektFactory = objektFactory;

        try {
            LOGGER.fine("Prüfe Datenbankstruktur");
            updateTables();
        } catch (GanglinienSpeicherException e) {
            throw new IllegalStateException(e);
        }

        long firstDelay = Math.abs(Duration.between(LocalTime.now(), compressTime).toMinutes());
        long intervalInMinuten = TimeUnit.DAYS.toMinutes(1);
        LOGGER.config("Starte tägliche Compressionstask in " + firstDelay + " Minuten, aller " + intervalInMinuten + " Minuten");
        Executors.newSingleThreadScheduledExecutor().scheduleAtFixedRate(() -> compress(), firstDelay,
                intervalInMinuten, TimeUnit.MINUTES);
    }

    private void compress() {

        synchronized (dbLock) {
            LOGGER.info(() -> "Compress task started");
            try (Connection connection = connect(); Statement statement = connection.createStatement()) {
                statement.execute("call SYSCS_UTIL.SYSCS_COMPRESS_TABLE('APP', 'GANGLINIEN', 1)");
                statement.execute("call SYSCS_UTIL.SYSCS_COMPRESS_TABLE('APP', 'STUETZSTELLEN', 1)");
                LOGGER.info(() -> "Compress task finished");
            } catch (GanglinienSpeicherException | SQLException e) {
                LOGGER.warning(() -> "Compress task failed: " + e.getLocalizedMessage());
            }
        }
    }

    public void dropNewTables() throws GanglinienSpeicherException {
        try (Connection connection = connect(); Statement statement = connection.createStatement()) {
            try {
                statement.execute("DROP TABLE APP.STUETZSTELLEN");
            } catch (SQLException ignore) {
                // ignore
            }
            try {
                statement.execute("DROP TABLE APP.GANGLINIEN");
            } catch (SQLException ignore) {
                // ignore
            }
        } catch (SQLException e) {
            throw new GanglinienSpeicherException(e);
        }
    }

    private void createTables(Connection connection) throws SQLException {

        synchronized (dbLock) {
            try (Statement statement = connection.createStatement()) {
                DatabaseMetaData metaData = connection.getMetaData();
                LOGGER.fine("Prüfe Tabelle GANGLINIEN");
                try (ResultSet tables = metaData.getTables(null, "APP", "GANGLINIEN", null)) {
                    if (!tables.next()) {
                        LOGGER.fine("Tabellen GANGLINIEN und STUETZSTELLEN anlegen");
                        statement.execute(
                                "CREATE TABLE APP.GANGLINIEN(\"ID\" BIGINT NOT NULL GENERATED ALWAYS AS IDENTITY (START WITH 1, INCREMENT BY 1) PRIMARY KEY, \"MQPID\" VARCHAR(128), \"EREIGNISTYPPID\" VARCHAR(128), \"REFERENZ\" BOOLEAN, \"VERSCHMELZUNGEN\" BIGINT, \"LETZTEVERSCHMELZUNG\" TIMESTAMP, \"TYP\" INTEGER, \"APPROXIMATION\" INTEGER, \"ORDNUNG\" INTEGER)");
                        statement.execute("CREATE INDEX GL_MQPID_INDEX ON GANGLINIEN (MQPID)");
                        statement.execute("CREATE INDEX GL_MQPID_ETPID_INDEX ON GANGLINIEN (MQPID,EREIGNISTYPPID)");
                        statement.execute(
                                "CREATE TABLE APP.STUETZSTELLEN(\"ID\" BIGINT REFERENCES APP.GANGLINIEN(\"ID\") ON DELETE CASCADE ON UPDATE RESTRICT, \"OFFSETMS\" BIGINT, \"QLKW\" DOUBLE, \"QKFZ\" DOUBLE, \"VPKW\" DOUBLE, \"VLKW\" DOUBLE)");
                        statement.execute("CREATE INDEX SST_ID_INDEX ON STUETZSTELLEN(ID)");
                    }
                }
            }
        }
    }

    private void updateTables() throws GanglinienSpeicherException {
        synchronized (dbLock) {
            try (Connection connection = connect()) {
                connection.setAutoCommit(false);
                createTables(connection);
                DatabaseMetaData metaData = connection.getMetaData();
                LOGGER.fine("Prüfe Tabelle DBMESSQUERSCHNITT");
                try (ResultSet tables = metaData.getTables(null, "APP", "DBMESSQUERSCHNITT", null)) {
                    if (tables.next()) {
                        runUpdate(connection);
                        try (Statement statement = connection.createStatement()) {
                            statement.execute("DROP TABLE APP.DBMESSQUERSCHNITT");
                        }
                    }
                }
                connection.commit();
            } catch (SQLException | ClassNotFoundException | IOException | GanglinienSpeicherException e) {
                throw new GanglinienSpeicherException(e);
            }
        }
    }

    private void runUpdate(Connection connection)
            throws SQLException, IOException, ClassNotFoundException, GanglinienSpeicherException {

        LOGGER.fine("Aktualisiere Gangliniendaten");

        String sqlStr = "SELECT MQPID,GANGLINIEN FROM APP.DBMESSQUERSCHNITT";
        try (PreparedStatement statement = connection.prepareStatement(sqlStr)) {
            try (ResultSet resultSet = statement.executeQuery()) {

                while (resultSet.next()) {
                    String mqPid = resultSet.getString("MQPID");
                    Blob ganglinienBlob = resultSet.getBlob("GANGLINIEN");
                    byte[] bytes = ganglinienBlob.getBytes(1L, (int) ganglinienBlob.length());
                    ObjectInputStream inputStream = new ObjectInputStream(new ByteArrayInputStream(bytes));

                    List<MqGanglinie> ganglinien = new ArrayList<>();

                    @SuppressWarnings("unchecked")
                    ArrayList<DbGanglinie> dbGanglinien = (ArrayList<DbGanglinie>) inputStream.readObject();
                    for (DbGanglinie dbGanglinie : dbGanglinien) {
                        MqGanglinie mqGanglinie = ermittleAlteMqGanglinie(dbGanglinie);
                        if (mqGanglinie != null) {
                            ganglinien.add(mqGanglinie);
                        }
                    }

                    LOGGER.info("DB-Aktualisierung: Sichere " + ganglinien.size() + " Ganglinien für MQ: " + mqPid);
                    sichereGanglinien(connection, ganglinien);
                }
            }
        }
    }

    static {
        try {
            Class.forName("org.apache.derby.jdbc.EmbeddedDriver");
        } catch (ClassNotFoundException e) {
            throw new IllegalStateException(e);
        }
    }

    private Connection connect() throws GanglinienSpeicherException {

        Connection connection;
        try {
            LOGGER.fine(() -> "Verbindung aufbauen: " + "jdbc:derby:" + dbDir + ";create=true");
            connection = DriverManager.getConnection("jdbc:derby:" + dbDir + ";create=true");
            LOGGER.fine(() -> "Verbindung hergestellt");
        } catch (SQLException e) {
            throw new GanglinienSpeicherException(e);
        }

        return connection;
    }

    @Override
    public List<MqGanglinie> leseGanglinien(MessQuerschnitt messQuerschnitt, Collection<EreignisTyp> ereignisTypen)
            throws GanglinienSpeicherException {

        synchronized (dbLock) {
            LOGGER.finer(() -> "Starte Lesen");
            List<MqGanglinie> ganglinien = new ArrayList<>();
            String sqlStr = "SELECT * FROM APP.GANGLINIEN WHERE MQPID = ?";
            try (Connection connection = connect();
                    PreparedStatement lesenStatement = connection.prepareStatement(sqlStr)) {
                lesenStatement.setString(1, messQuerschnitt.getPid());
                try (ResultSet resultSet = lesenStatement.executeQuery()) {

                    while (resultSet.next()) {
                        long key = resultSet.getLong("ID");
                        MqGanglinie mqGanglinie = new MqGanglinie(key,
                                MessQuerschnitt.of(resultSet.getString("MQPID")));
                        mqGanglinie.setAnzahlVerschmelzungen(resultSet.getLong("VERSCHMELZUNGEN"));
                        mqGanglinie.setApproximationDaK(ApproximationTyp.with(resultSet.getInt("APPROXIMATION")));
                        mqGanglinie.setBSplineOrdnung(resultSet.getInt("ORDNUNG"));
                        mqGanglinie.setEreignisTyp(EreignisTyp.of(resultSet.getString("EREIGNISTYPPID")));
                        mqGanglinie.setLetzteVerschmelzung(LocalDateTimeKonverter
                                .fromMillis(resultSet.getTimestamp("LETZTEVERSCHMELZUNG").getTime()));
                        mqGanglinie.setReferenz(resultSet.getBoolean("REFERENZ"));
                        mqGanglinie.setTyp(GanglinieTyp.with(resultSet.getInt("TYP")));

                        if (ereignisTypen.isEmpty() || ereignisTypen.contains(mqGanglinie.getEreignisTyp())) {
                            readStuetzStellen(connection, key, mqGanglinie);
                            ganglinien.add(mqGanglinie);
                        }
                    }
                }
            } catch (SQLException e) {
                throw new GanglinienSpeicherException(e);
            }
            LOGGER.finer(() -> "Beende Lesen");
            return ganglinien;
        }
    }

    private void readStuetzStellen(Connection connection, long key, MqGanglinie mqGanglinie) throws SQLException {
        String sqlStr = "SELECT * FROM APP.STUETZSTELLEN WHERE ID = ?";

        try (PreparedStatement lesenStatement = connection.prepareStatement(sqlStr)) {
            lesenStatement.setLong(1, key);
            try (ResultSet resultSet = lesenStatement.executeQuery()) {

                while (resultSet.next()) {

                    Double qKfz = null;
                    double wert = resultSet.getDouble("QKFZ");
                    if (Double.compare(wert, Messwerte.UNDEFINIERT) != 0) {
                        qKfz = wert;
                    }

                    Double qLkw = null;
                    wert = resultSet.getDouble("QLKW");
                    if (Double.compare(wert, Messwerte.UNDEFINIERT) != 0) {
                        qLkw = wert;
                    }

                    Double vPkw = null;
                    wert = resultSet.getDouble("VPKW");
                    if (Double.compare(wert, Messwerte.UNDEFINIERT) != 0) {
                        vPkw = wert;
                    }

                    Double vLkw = null;
                    wert = resultSet.getDouble("VLKW");
                    if (Double.compare(wert, Messwerte.UNDEFINIERT) != 0) {
                        vLkw = wert;
                    }

                    Messwerte messWert = new Messwerte(qKfz, qLkw, vPkw, vLkw);
                    mqGanglinie.aktualisiereMesswert(resultSet.getLong("OFFSETMS"), messWert);
                }
            }
        }
    }

    @Override
    public List<MqGanglinie> schreibeGanglinien(MessQuerschnitt messQuerschnitt, List<MqGanglinie> ganglinien)
            throws GanglinienSpeicherException {

        List<MqGanglinie> vorherigeGanglinien = leseGanglinien(messQuerschnitt, Collections.emptyList());
        Set<Long> vorherigeGanglinienId = new LinkedHashSet<>();
        vorherigeGanglinien.forEach(gl -> vorherigeGanglinienId.add(gl.getGanglinieId()));

        synchronized (dbLock) {
            try (Connection connection = connect()) {
                connection.setAutoCommit(false);
                List<MqGanglinie> result = sichereGanglinien(connection, ganglinien);
                result.forEach(gl -> vorherigeGanglinienId.remove(gl.getGanglinieId()));
                if (!vorherigeGanglinienId.isEmpty()) {
                    loescheGanglinienMitId(connection, messQuerschnitt, vorherigeGanglinienId);
                }
                connection.commit();
                return result;
            } catch (SQLException e) {
                throw new GanglinienSpeicherException(e);
            }
        }
    }

    private List<MqGanglinie> sichereGanglinien(Connection connection, List<MqGanglinie> ganglinien)
            throws GanglinienSpeicherException {

        List<MqGanglinie> result = new ArrayList<>();

        try {
            for (MqGanglinie mqGanglinie : ganglinien) {
                if (mqGanglinie.getGanglinieId() > 0) {
                    result.add(aktualisiereGanglinie(connection, mqGanglinie));
                } else {
                    result.add(erzeugeNeueGanglinie(connection, mqGanglinie));
                }
            }
        } catch (SQLException e) {
            throw new GanglinienSpeicherException(e);
        }
        return result;
    }

    private MqGanglinie erzeugeNeueGanglinie(Connection connection, MqGanglinie ganglinie)
            throws GanglinienSpeicherException {

        String insGlStr = "INSERT INTO APP.GANGLINIEN (\"MQPID\", \"EREIGNISTYPPID\", \"REFERENZ\", \"VERSCHMELZUNGEN\", \"LETZTEVERSCHMELZUNG\", \"TYP\", \"APPROXIMATION\", \"ORDNUNG\") VALUES(?,?,?,?,?,?,?,?)";
        try (PreparedStatement insStatement = connection.prepareStatement(insGlStr, Statement.RETURN_GENERATED_KEYS)) {

            MessQuerschnitt mq = ganglinie.getMessQuerschnitt();

            insStatement.setString(1, mq.getPid());
            insStatement.setString(2, ganglinie.getEreignisTyp().getPid());
            insStatement.setBoolean(3, ganglinie.isReferenz());
            insStatement.setLong(4, ganglinie.getAnzahlVerschmelzungen());
            insStatement.setTimestamp(5,
                    new Timestamp(LocalDateTimeKonverter.toMillis(ganglinie.getLetzteVerschmelzung())));
            insStatement.setInt(6, ganglinie.getTyp().getCode());
            insStatement.setInt(7, ganglinie.getApproximationDaK().getCode());
            insStatement.setInt(8, ganglinie.getBSplineOrdnung());
            insStatement.execute();

            long key = 0;
            try (ResultSet keys = insStatement.getGeneratedKeys()) {
                if (keys.next()) {
                    key = keys.getLong(1);
                    schreibeStuetzStellen(connection, key, ganglinie.getStuetzstellen(), false);
                }
            }

            if (key > 0) {
                return ganglinie.erzeugeKopieMitId(key);
            }

        } catch (SQLException e) {
            throw new GanglinienSpeicherException(e);
        }

        throw new GanglinienSpeicherException("ID für neue Ganglinie konnte nicht bestimmt werden.");
    }

    private void schreibeStuetzStellen(Connection connection, long key, List<Stuetzstelle<Messwerte>> stuetzstellen,
            boolean delete) throws SQLException {
        String delStSt = "DELETE FROM APP.STUETZSTELLEN WHERE ID=?";
        String insStSt = "INSERT INTO APP.STUETZSTELLEN(\"ID\",\"OFFSETMS\",\"QLKW\",\"QKFZ\",\"VPKW\",\"VLKW\") VALUES(?,?,?,?,?,?)";
        try (PreparedStatement stStatment = connection.prepareStatement(insStSt);
                PreparedStatement delStatment = connection.prepareStatement(delStSt)) {

            if (delete) {
                delStatment.setLong(1, key);
                delStatment.execute();
            }

            for (Stuetzstelle<Messwerte> ss : stuetzstellen) {
                stStatment.setLong(1, key);
                stStatment.setLong(2, ss.getZeitstempel());
                stStatment.setDouble(3, ss.getWert().getQLkwNichtNull());
                stStatment.setDouble(4, ss.getWert().getQKfzNichtNull());
                stStatment.setDouble(5, ss.getWert().getVPkwNichtNull());
                stStatment.setDouble(6, ss.getWert().getVLkwNichtNull());
                stStatment.execute();
            }
        }
    }

    private MqGanglinie aktualisiereGanglinie(Connection connection, MqGanglinie ganglinie) throws SQLException {

        String updGlStr = "UPDATE APP.GANGLINIEN SET \"MQPID\"=?, \"EREIGNISTYPPID\"=?, \"REFERENZ\"=?, \"VERSCHMELZUNGEN\"=?, \"LETZTEVERSCHMELZUNG\"=?, \"TYP\"=?, \"APPROXIMATION\"=?, \"ORDNUNG\"=? WHERE \"ID\"=?";
        try (PreparedStatement statement = connection.prepareStatement(updGlStr, Statement.NO_GENERATED_KEYS)) {

            MessQuerschnitt mq = ganglinie.getMessQuerschnitt();

            statement.setString(1, mq.getPid());
            statement.setString(2, ganglinie.getEreignisTyp().getPid());
            statement.setBoolean(3, ganglinie.isReferenz());
            statement.setLong(4, ganglinie.getAnzahlVerschmelzungen());
            statement.setTimestamp(5,
                    new Timestamp(LocalDateTimeKonverter.toMillis(ganglinie.getLetzteVerschmelzung())));
            statement.setInt(6, ganglinie.getTyp().getCode());
            statement.setInt(7, ganglinie.getApproximationDaK().getCode());
            statement.setInt(8, ganglinie.getBSplineOrdnung());
            statement.setLong(9, ganglinie.getGanglinieId());
            statement.execute();

            schreibeStuetzStellen(connection, ganglinie.getGanglinieId(), ganglinie.getStuetzstellen(), true);
            return ganglinie;
        }
    }

    @Override
    public void loescheGanglinien(MessQuerschnitt messQuerschnitt, Collection<EreignisTyp> ereignisTypen,
            Collection<Long> ganglinienId) throws GanglinienSpeicherException {

        synchronized (dbLock) {
            try (Connection connection = connect()) {
                connection.setAutoCommit(false);
                if (!ganglinienId.isEmpty()) {
                    loescheGanglinienMitId(connection, messQuerschnitt, ganglinienId);
                } else if (!ereignisTypen.isEmpty()) {
                    loescheGanglinienMitEreignisTypen(connection, messQuerschnitt, ereignisTypen);
                } else {

                    String sqlStr = "DELETE FROM APP.GANGLINIEN WHERE MQPID = ?";
                    try (PreparedStatement loeschenStatement = connection.prepareStatement(sqlStr)) {
                        connection.setAutoCommit(false);
                        loeschenStatement.setString(1, messQuerschnitt.getPid());
                        loeschenStatement.execute();
                        connection.commit();
                    }
                }
                connection.commit();
            } catch (SQLException e) {
                throw new GanglinienSpeicherException(e);
            }
        }
    }

    private void loescheGanglinienMitEreignisTypen(Connection connection, MessQuerschnitt messQuerschnitt,
            Collection<EreignisTyp> ereignisTypen) throws GanglinienSpeicherException {

        String sqlStr = "DELETE FROM APP.GANGLINIEN WHERE MQPID = ? AND EREIGNISTYPPID = ?";
        try (PreparedStatement loeschenStatement = connection.prepareStatement(sqlStr)) {

            for (EreignisTyp ereignisTyp : ereignisTypen) {
                loeschenStatement.setString(1, messQuerschnitt.getPid());
                loeschenStatement.setString(2, ereignisTyp.getPid());
                loeschenStatement.execute();
            }
        } catch (SQLException e) {
            throw new GanglinienSpeicherException(e);
        }

    }

    private void loescheGanglinienMitId(Connection connection, MessQuerschnitt messQuerschnitt,
            Collection<Long> ganglinienId) throws GanglinienSpeicherException {

        String sqlStr = "DELETE FROM APP.GANGLINIEN WHERE MQPID = ? AND ID = ?";
        try (PreparedStatement loeschenStatement = connection.prepareStatement(sqlStr)) {
            for (long id : ganglinienId) {
                loeschenStatement.setString(1, messQuerschnitt.getPid());
                loeschenStatement.setLong(2, id);
                loeschenStatement.execute();
            }
        } catch (SQLException e) {
            throw new GanglinienSpeicherException(e);
        }
    }

    private MqGanglinie ermittleAlteMqGanglinie(DbGanglinie dbGanglinie) {

        MessQuerschnitt mq = objektFactory.createMessQuerschnitt(dbGanglinie.getMqPid());
        if (mq == null) {
            LOGGER.warning(() -> "Es konnte kein MQ mit der PID " + dbGanglinie.getMqPid() + " ermittelt werden.");
            mq = MessQuerschnitt.of(dbGanglinie.getMqPid());
        }

        EreignisTyp ereignisTyp = null;
        if (objektFactory instanceof DavObjektFactory) {
            SystemObject so = ((DavObjektFactory) objektFactory).getDav().getDataModel()
                    .getObject(dbGanglinie.getEreignisTypId());
            if (so != null) {
                ereignisTyp = objektFactory.createEreignisTyp(so.getPid());
            }
        }

        if (ereignisTyp == null) {
            LOGGER.warning(() -> "Es konnte kein EreignisTyp mit der ID " + dbGanglinie.getEreignisTypId()
                    + " ermittelt werden.");
            return null;
        }

        return dbGanglinie.convert(mq, ereignisTyp);
    }

    public void close() {
        try (Connection connection = DriverManager.getConnection("jdbc:derby:;shutdown=true")) {
            LOGGER.fine("Datenbank herunterfahren");
            // nothing to do
        } catch (SQLException e) {
            LOGGER.fine(() -> "Datenbank geschlossen mit: " + e.getLocalizedMessage());
        }
        LOGGER.fine(() -> "Datenbank geschlossen");
    }

    @Override
    public List<MqGanglinie> aktualisiereGanglinien(List<MqGanglinie> ganglinien) throws GanglinienSpeicherException {

        synchronized (dbLock) {
            List<MqGanglinie> result = new ArrayList<>();
            try (Connection connection = connect()) {
                connection.setAutoCommit(false);
                for (MqGanglinie mqGanglinie : ganglinien) {
                    if (mqGanglinie.getGanglinieId() > 0) {
                        result.add(aktualisiereGanglinie(connection, mqGanglinie));
                    } else {
                        throw new GanglinienSpeicherException("Eine zu aktualisierende Ganglinie hat keine Id");
                    }
                }
                connection.commit();
            } catch (SQLException e) {
                throw new GanglinienSpeicherException(e);
            }
            return result;
        }
    }

    @Override
    public List<MqGanglinie> erzeugeGanglinien(List<MqGanglinie> ganglinien) throws GanglinienSpeicherException {

        synchronized (dbLock) {
            List<MqGanglinie> result = new ArrayList<>();
            try (Connection connection = connect()) {
                connection.setAutoCommit(false);
                for (MqGanglinie mqGanglinie : ganglinien) {
                    result.add(erzeugeNeueGanglinie(connection, mqGanglinie));
                }
                connection.commit();
            } catch (SQLException e) {
                throw new GanglinienSpeicherException(e);
            }
            return result;
        }
    }

    @Override
    public int zaehleStuetzstellen(long ganglinieId) throws GanglinienSpeicherException {

        String sqlStr = "SELECT COUNT(*) FROM APP.STUETZSTELLEN WHERE ID=?";

        try (Connection connection = connect(); PreparedStatement statement = connection.prepareStatement(sqlStr)) {
            statement.setLong(1, ganglinieId);
            try (ResultSet resultSet = statement.executeQuery()) {
                if (resultSet.next()) {
                    return resultSet.getInt(1);
                }
            }
        } catch (SQLException e) {
            throw new GanglinienSpeicherException(e);
        }

        return -1;
    }
}
